(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory DP_Transformation
imports
  QDP_Framework
  Lexicographic_Extension
begin

fun
  initial_dpp :: "('f \<Rightarrow> 'f) \<Rightarrow> bool \<Rightarrow> bool \<Rightarrow> ('f, 'v) term set \<Rightarrow> ('f, 'v) trs  \<Rightarrow> ('f, 'v) dpp"
where
  "initial_dpp shp nfs m Q R = (nfs, m, DP shp (applicable_rules Q R), {}, Q, {}, R)"


(* @{term "Tinf"} characterizes termination. *)
lemma Tinf_SN_conv: "(Tinf R = {}) = (SN R)"
proof
  assume "Tinf R = {}" thus "SN R" using not_SN_imp_Tinf by blast
next
  assume "SN R" thus "Tinf R = {}" unfolding Tinf_def SN_defs by blast
qed

definition "wwf_inn_qtrs nfs Q R \<equiv> nfs \<and> NF_terms Q \<subseteq> NF_trs R \<and> Ball (fst ` R) is_Fun"

lemma wwf_inn_qtrs_imp_left_fun: 
  "wwf_qtrs Q R \<or> wwf_inn_qtrs nfs Q R \<Longrightarrow> (l,r) \<in> R \<Longrightarrow> is_Fun l"
  using wwf_qtrs_imp_left_fun[of Q R l r] unfolding wwf_inn_qtrs_def by auto

(* TODO: move to abstract rewriting *)
lemma NF_imp_SN_on: "a \<in> NF r \<Longrightarrow> SN_on r {a}"
  by blast

lemma Tinf_lemma_1:
  assumes wwf: "wwf_qtrs Q R \<or> wwf_inn_qtrs nfs Q R"
    and Tinf: "Fun f ts \<in> Tinf (qrstep nfs Q R)" (is "?t \<in> _")
  shows "\<exists>l r u \<sigma>.
  (l, r) \<in> applicable_rules Q R \<and> 
  r \<unrhd> u \<and> 
  is_Fun u \<and> 
  (?t, l \<cdot> \<sigma>) \<in> (nrqrstep nfs Q R)^* \<and> 
  (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> rqrstep nfs Q R \<and> 
  r \<cdot> \<sigma> \<unrhd> u \<cdot> \<sigma> \<and>
  l \<cdot> \<sigma> \<in> Tinf (qrstep nfs Q R) \<and>
  u \<cdot> \<sigma> \<in> Tinf (qrstep nfs Q R) \<and>
  \<not> (l \<rhd> u) \<and>
  NF_subst nfs (l,r) \<sigma> Q"
proof -
  let ?U = "applicable_rules Q R"
  let ?qr = "qrstep nfs Q R"
  let ?rqr = "rqrstep nfs Q R"
  let ?nrqr = "nrqrstep nfs Q R"
  from Tinf_imp_SN_nr_first_root_step[OF Tinf]
  obtain lhs rhs where nsteps: "(Fun f ts, lhs) \<in> ?nrqr^*" 
    and step: "(lhs,rhs) \<in> ?rqr" and nSN: "\<not> SN_on ?qr {rhs}" by auto
  then obtain l r \<sigma>
    where NF: "\<forall>v\<lhd>l \<cdot> \<sigma>. v \<in> NF_terms Q"
    and "(l, r) \<in> R" and lhs: "l \<cdot> \<sigma> = lhs" and rhs: "r \<cdot> \<sigma> = rhs" 
    and nfs: "NF_subst nfs (l, r) \<sigma> Q" by auto
  with step have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> ?rqr" by auto
  from `(l, r) \<in> R` only_applicable_rules[OF NF] have "(l, r) \<in> ?U"
    unfolding applicable_rules_def by auto
  from nrqrsteps_preserve_root_fun[OF nsteps]
  obtain ls' where "lhs = Fun f ls'" by auto
  have A: "\<And>i. (ts ! i, ls' ! i) \<in> ?qr^*" (is "\<And>i. ?P i")
  proof -
    fix j
    from nsteps have "(args ?t!j, args (Fun f ls')!j) \<in> ?qr^*"
      unfolding `lhs = Fun f ls'` by (rule nrqrsteps_imp_arg_qrsteps)
    thus "?P j" by simp
  qed
  from nsteps and `lhs = Fun f ls'` have "(?t,Fun f ls') \<in> ?nrqr^*" by simp
  from `l\<cdot>\<sigma> = lhs` and `lhs = Fun f ls'` have "l\<cdot>\<sigma> = Fun f ls'" by simp
  with `(?t,Fun f ls') \<in> ?nrqr^*` have "(?t,l\<cdot>\<sigma>) \<in> ?nrqr^*" by simp
  from `(?t,Fun f ls') \<in> ?nrqr^*` have numargs: "num_args ?t = num_args (Fun f ls')"
    by (rule nrqrsteps_num_args)
  from wwf_inn_qtrs_imp_left_fun[OF wwf `(l,r) \<in> R`] obtain g ls where "l = Fun g ls" by force
  hence "l\<cdot>\<sigma> = Fun g (map (\<lambda>t. t\<cdot>\<sigma>) ls)" by simp
  with `l\<cdot>\<sigma> = Fun f ls'` have "g = f" and "ls' = map (\<lambda>t. t\<cdot>\<sigma>) ls" by auto
  hence B: "\<forall>i<length ls'. ls' ! i = (ls ! i) \<cdot> \<sigma>" by auto
  have "\<forall>i < num_args ?t. (ts ! i, (ls ! i) \<cdot> \<sigma>) \<in> ?qr^*"
  proof (intro allI impI)
    fix i assume "i < num_args (Fun f ts)"
    with numargs have "i < length ls'" by simp
    with A[of i] show "(ts ! i, ls ! i \<cdot> \<sigma>) \<in> ?qr^*" using B by simp
  qed
  from `l \<cdot> \<sigma> = Fun f ls'` numargs have "num_args ?t = num_args (l\<cdot>\<sigma>)"  by simp
  with `l\<cdot>\<sigma> = Fun f ls'` and `\<forall>i<length ls'. ls'!i = (ls!i)\<cdot>\<sigma>`
  have "!!i. i < num_args ?t \<Longrightarrow> (ls!i)\<cdot>\<sigma> = ls'!i" by auto
  have "!!i. i < num_args ?t \<Longrightarrow> ?t \<rhd> (ts!i)"
  proof -
    fix i assume "i < num_args ?t"
    hence "ts!i \<in> set ts" by simp
    thus "?t \<rhd> (ts!i)" by (rule supt.arg)
  qed
  with Tinf[unfolded Tinf_def] have "!!i. i < num_args ?t \<Longrightarrow> SN_on ?qr {ts ! i}" by auto
  with `\<forall>i<num_args ?t. (ts!i,(ls!i)\<cdot>\<sigma>) \<in> ?qr^*` 
  have "!!i. i < num_args ?t \<Longrightarrow> SN_on ?qr {(ls!i) \<cdot> \<sigma>}" using steps_preserve_SN_on[of _ _ ?qr] by blast
  with `!!i. i < num_args ?t \<Longrightarrow> (ls!i)\<cdot>\<sigma> = ls'!i` and `num_args ?t = num_args (l\<cdot>\<sigma>)`
    and `l\<cdot>\<sigma> = Fun f ls'` and `!!i. i < num_args ?t \<Longrightarrow> SN_on ?qr {(ls!i) \<cdot> \<sigma>}`
  have argsSN: "!!i. i < num_args (Fun f ls') \<Longrightarrow> SN_on ?qr {ls' ! i}" by auto
  have "\<forall>s\<lhd>Fun f ls'. SN_on ?qr {s}" 
  proof (intro allI impI)
    fix s
    assume tmp: "Fun f ls' \<rhd> s"
    then obtain C where "C \<noteq> \<box>" and "Fun f ls' = C\<langle>s\<rangle>" by auto
    then obtain bef aft C where id: "Fun f ls' = Fun f (bef @ C\<langle>s\<rangle> # aft)" by (cases C, auto)
    hence len: "length bef < num_args (Fun f ls')" by auto
    with argsSN[OF len] id have "SN_on ?qr {C \<langle>s\<rangle>}" by auto
    with restrict_SN_imp_SN_subt[OF ctxt_closed_qrstep _ ctxt_imp_supteq[of C s]]
    show "SN_on ?qr {s}" by auto
  qed
  hence "\<forall>s\<lhd>l\<cdot>\<sigma>. SN_on ?qr {s}" unfolding `l\<cdot>\<sigma> = Fun f ls'` by simp 
  from nSN rhs obtain B where B0: "B 0 = r \<cdot> \<sigma>" and Bsteps: "\<And> i. (B i, B (Suc i)) \<in> ?qr" by auto
  from `(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> ?rqr` have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> ?qr" 
    unfolding qrstep_iff_rqrstep_or_nrqrstep by auto
  let ?B = "\<lambda>j. case j of 0 \<Rightarrow> l \<cdot> \<sigma> | Suc j \<Rightarrow> B j"
  have "?B 0 = l\<cdot>\<sigma>" using lhs by simp
  {
    fix i
    have "(?B i,?B(Suc i)) \<in> ?qr" using Bsteps[of "i - Suc 0"] `(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> ?qr` B0 by (cases i, auto)
  } note Bsteps = this
  with `?B 0 = l\<cdot>\<sigma>` have "\<not> SN_on ?qr {l \<cdot> \<sigma>}" unfolding SN_on_def by best
  with `\<forall>s\<lhd>l\<cdot>\<sigma>. SN_on ?qr {s}` have tinf_ls: "l \<cdot> \<sigma> \<in> Tinf ?qr" unfolding Tinf_def by simp
  {
    fix x
    assume x: "x \<in> vars_term r"
    from wwf have "SN_on ?qr {Var x \<cdot> \<sigma>}"
    proof
      assume "wwf_qtrs Q R"
      with x have "x \<in> vars_term l" using `(l,r) \<in> ?U` unfolding wwf_qtrs_def wwf_rule_def applicable_rules_def by auto
      hence "l \<rhd> Var x" using `l = Fun g ls` supteq_Var[OF `x \<in> vars_term l`]
        unfolding supt_supteq_conv by auto
      from supt_stable[OF this, of \<sigma>] tinf_ls show ?thesis unfolding Tinf_def by auto
    next
      assume "wwf_inn_qtrs nfs Q R"
      with nfs x have "\<sigma> x \<in> NF_trs R" unfolding NF_subst_def wwf_inn_qtrs_def vars_rule_def by auto
      hence "SN_on (rstep R) {\<sigma> x}" by (rule NF_imp_SN_on)
      from SN_on_mono[OF this]
      show ?thesis by auto
    qed
  } note xSN = this
  from nSN rhs have "\<not> SN_on ?qr {r \<cdot> \<sigma>}" by simp
  then obtain t' where "r\<cdot>\<sigma> \<unrhd> t'" and "t' \<in> Tinf ?qr" using not_SN_imp_subt_Tinf by auto
  have "\<forall>x \<in> vars_term r. \<not>((Var x)\<cdot>\<sigma> \<unrhd> t')"
  proof
    fix x assume "x \<in> vars_term r"
    hence "SN_on ?qr {Var x \<cdot> \<sigma>}" using xSN by simp
    show "\<not> Var x\<cdot>\<sigma> \<unrhd> t'"
    proof
      assume "Var x\<cdot>\<sigma> \<unrhd> t'"
      with `SN_on ?qr {Var x \<cdot> \<sigma>}` have "SN_on ?qr {t'}" using restrict_SN_imp_SN_subt[OF ctxt_closed_qrstep] by auto
      with `t' \<in> Tinf ?qr` show "False" unfolding Tinf_def by auto
    qed
  qed
  with `r\<cdot>\<sigma> \<unrhd> t'` have "\<exists>u. r \<unrhd> u \<and> t' = u\<cdot>\<sigma>" using subt_instance_and_not_subst_imp_subt[where s = r and \<sigma> = \<sigma>] by auto
  then obtain u where "r \<unrhd> u" and "t' = u\<cdot>\<sigma>" by best
  have "is_Fun u"
  proof (rule ccontr)
    assume "\<not> is_Fun u"
    hence "\<exists>x. u = Var x" by (cases u) simp_all
    then obtain x where "u = Var x" by best
    with `r \<unrhd> u` have "r \<unrhd> Var x" by auto
    hence "x \<in> vars_term r" using var_subt_imp_subt_in_vars by best
    with xSN `u = Var x` have "SN_on ?qr {u \<cdot> \<sigma>}" by simp
    with `t' = u\<cdot>\<sigma>` have "SN_on ?qr {t'}" by simp
    with `t' \<in> Tinf ?qr` show "False" unfolding Tinf_def by simp
  qed
  from `t' = u\<cdot>\<sigma>` and `t' \<in> Tinf ?qr` have "u \<cdot> \<sigma> \<in> Tinf ?qr" by simp
  have "\<not> l \<rhd> u"
  proof
    assume "l \<rhd> u"
    hence "l \<cdot> \<sigma> \<rhd> u \<cdot> \<sigma>" by (rule supt_stable)
    with `l \<cdot> \<sigma> \<in> Tinf ?qr` have "SN_on ?qr {u \<cdot> \<sigma>}" unfolding Tinf_def by best
    with `u \<cdot> \<sigma> \<in> Tinf ?qr` show "False" unfolding Tinf_def by auto
  qed
  from `r \<unrhd> u` have "r\<cdot>\<sigma> \<unrhd> (u\<cdot>\<sigma>)" by (rule supteq_stable)
  from `(l,r) \<in> ?U` and `r \<unrhd> u` and `is_Fun u` and `(?t,l\<cdot>\<sigma>) \<in> ?nrqr^*` 
    `(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> ?rqr` and `u \<cdot> \<sigma> \<in> Tinf ?qr` and `r \<cdot> \<sigma> \<unrhd> u \<cdot> \<sigma>` and `l \<cdot> \<sigma> \<in> Tinf ?qr`
    and `\<not> l \<rhd> u` nfs
  show ?thesis by blast
qed

lemma Tinf_var:
  assumes "wwf_qtrs Q R \<or> wwf_inn_qtrs nfs Q R" shows "Var x \<notin> Tinf (qrstep nfs Q R)"
  using SN_on_Var_gen[of R nfs Q x] wwf_inn_qtrs_imp_left_fun[OF assms] unfolding Tinf_def by force
 
lemma Tinf_lemma:
  assumes wwf: "wwf_qtrs Q R \<or> wwf_inn_qtrs nfs Q R" and tinf: "t\<in> Tinf (qrstep nfs Q R)"
  shows "\<exists>l r u \<sigma>.
    (l, r) \<in> applicable_rules Q R \<and> r \<unrhd> u \<and> is_Fun u \<and> (t, l \<cdot> \<sigma>) \<in> (nrqrstep nfs Q R)^*
    \<and> (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> rqrstep nfs Q R \<and> r \<cdot> \<sigma> \<unrhd> u \<cdot> \<sigma> \<and> l \<cdot> \<sigma> \<in> Tinf (qrstep nfs Q R)
    \<and> u \<cdot> \<sigma> \<in> Tinf (qrstep nfs Q R) \<and> \<not> (l \<rhd> u) \<and> NF_subst nfs (l,r) \<sigma> Q"
proof (cases t)
  case (Var x)
  with Tinf_var[OF wwf] tinf have False by auto
  thus ?thesis ..
next
  case (Fun f ts)
  from Tinf_lemma_1[OF wwf tinf[simplified Fun]]
  show ?thesis unfolding Fun .
qed

lemma Tinf_imp_def_root_qrstep:
  assumes wwf: "wwf_qtrs Q R \<or> wwf_inn_qtrs nfs Q R" and tinf: "t \<in> Tinf (qrstep nfs Q R)"
  shows "\<exists>f ss. t = Fun f ss \<and> defined (applicable_rules Q R) (f, length ss)"
proof -
  from Tinf_var[OF wwf] tinf obtain f ts where t: "t = Fun f ts" by (cases t, auto)
  with tinf have tinf: "Fun f ts \<in> Tinf (qrstep nfs Q R)" by auto
  from Tinf_lemma[OF wwf tinf]
  obtain l r \<sigma> where lr: "(l,r) \<in> applicable_rules Q R" and "(Fun f ts,l\<cdot>\<sigma>) \<in> (nrqrstep nfs Q R)^*" 
    by blast
  from lr have "(l,r) \<in> R" unfolding applicable_rules_def by auto
  from wwf_inn_qtrs_imp_left_fun[OF wwf this]
  obtain g ls where l: "l = Fun g ls" by blast
  with nrqrsteps_preserve_root[OF `(Fun f ts,l\<cdot>\<sigma>) \<in> (nrqrstep nfs Q R)^*`]
    nrqrsteps_num_args[OF `(Fun f ts,l\<cdot>\<sigma>) \<in> (nrqrstep nfs Q R)^*`]
  have "g = f" and "length ls = length ts" by auto
  with lr l have "defined (applicable_rules Q R) (f, length ts)" unfolding defined_def by auto
  with t show ?thesis by auto
qed

lemma Tinf_imp_Fun: 
   assumes wwf: "wwf_qtrs Q R \<or> wwf_inn_qtrs nfs Q R" and "t \<in> Tinf (qrstep nfs Q R)"
   shows "\<exists>f ts. t = Fun f ts"
   using Tinf_var[OF wwf] assms(2) by (cases t, auto)

definition wf_qtrs :: "bool \<Rightarrow> ('f, 'v) term set \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool" where
  "wf_qtrs nfs Q R \<equiv> (wwf_qtrs Q R \<or> wwf_inn_qtrs nfs Q R) \<and> (\<forall>t\<in>Q. is_Fun t)"

lemma wf_qtrs_wf_trs_conv: "wf_qtrs nfs {} = wf_trs"
proof (rule ext)
  fix R :: "('f,'v)trs"
  show "wf_qtrs nfs {} R = wf_trs R"
    unfolding wf_qtrs_def wf_trs_def wwf_qtrs_def wwf_inn_qtrs_def
    by (force simp: applicable_rule_empty)
qed

lemma Tinf_lemma_DP:
  assumes wf: "wf_qtrs nfs Q R" and s: "s \<in> Tinf (qrstep nfs Q R)"
    and ndef:
    "\<And>f n ss. defined (applicable_rules Q R) (f, n) \<Longrightarrow> length ss = n \<Longrightarrow> Fun (shp f) ss \<notin> Q"
  shows "\<exists>l r u \<sigma>. (l,r) \<in> DP shp (applicable_rules Q R) \<and> r \<cdot> \<sigma> = sharp_term shp u
    \<and> u \<in> Tinf (qrstep nfs Q R) \<and> (sharp_term shp s, l \<cdot> \<sigma>) \<in> (nrqrstep nfs Q R)^*
    \<and> l \<cdot> \<sigma> \<in> NF_terms Q \<and> NF_subst True (l,r) \<sigma> Q"
proof -
  let ?qr = "qrstep nfs Q R"
  let ?U = "applicable_rules Q R"
  let ?nrqr = "nrqrstep nfs Q R"
  let ?rqr = "rqrstep nfs Q R"
  from wf have wf: "wwf_qtrs Q R \<or> wwf_inn_qtrs nfs Q R" and wfQ: "\<forall>t\<in>Q. is_Fun t"
    unfolding wf_qtrs_def by auto
  let ?Q = "Q" and ?R = "R"
  let ?Iqr = "nrqrstep nfs ?Q ?R"
  obtain l r u \<sigma> where lr: "(l,r) \<in> ?U" and "r \<unrhd> u" and "is_Fun u" and "(s,l\<cdot>\<sigma>) \<in> ?nrqr^*"
    and step: "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> ?rqr" and "l \<cdot> \<sigma> \<in> Tinf ?qr" and "u \<cdot> \<sigma> \<in> Tinf ?qr" and "\<not> l \<rhd> u"
    and nfs: "NF_subst nfs (l,r) \<sigma> Q"
    using Tinf_lemma[OF wf s] by auto
  from lr have "(l,r) \<in> R" unfolding applicable_rules_def by auto
  let ?u = "u\<cdot>\<sigma>"
  let ?l = "sharp_term shp l"
  let ?r = "sharp_term shp u"
  let ?\<sigma> = \<sigma>
  obtain f us where "u = Fun f us" using `is_Fun u` by (cases u) auto
  have "root u = root ?u" unfolding `u = Fun f us` by simp
  obtain h vs where "?u = Fun h vs" and "defined ?U (h, length vs)"
    using Tinf_imp_def_root_qrstep[OF wf `?u \<in> Tinf ?qr`] by auto
  from `(s,l\<cdot>\<sigma>) \<in> ?nrqr^*` have "(sharp_term shp s, sharp_term shp (l \<cdot> \<sigma>)) \<in> ?Iqr^*"
    by (rule nrqrsteps_imp_sharp_qrsteps)
  from wwf_inn_qtrs_imp_left_fun[OF wf `(l,r) \<in> R`] obtain g ts where l: "l = Fun g ts" by force
  have "sharp_term shp (l\<cdot>\<sigma>) = ?l\<cdot>?\<sigma>" unfolding `l = Fun g ts` sharp_term_subst_distr by simp
  with `(sharp_term shp s,sharp_term shp (l\<cdot>\<sigma>)) \<in> ?Iqr^*` have "(sharp_term shp s,?l\<cdot>?\<sigma>) \<in> ?Iqr^*" by simp
  from `(u\<cdot>\<sigma>) = Fun h vs` and `u = Fun f us` have "f = h" and "length us = length vs" by auto
  with `defined ?U (h, length vs)` have "defined ?U (f, length us)" by simp
  from `(l,r) \<in> ?U` and `r \<unrhd> u` and `defined ?U (f, length us)` and `\<not>(l \<rhd> u)` have "(?l,?r) \<in> DP shp ?U" unfolding DP_def `u = Fun f us` by auto
  have "?r\<cdot>?\<sigma> = sharp_term shp ?u" unfolding `u = Fun f us` sharp_term_subst_distr by simp
  from step have NF: "\<forall>u\<lhd>l\<cdot>\<sigma>. u \<in> NF_terms Q" unfolding rqrstep_def rstep_r_c_s_def by auto
  from lr l have gdef: "defined ?U (g, length ts)" unfolding defined_def by auto
  note ndef = ndef[OF gdef]
  note NF = NF[unfolded l]
  have "?l \<cdot> ?\<sigma> \<in> NF_terms ?Q"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    from not_NF_termsE[OF this]
    obtain q C \<gamma> where q: "q \<in> ?Q" and match: "?l \<cdot> ?\<sigma> = C\<langle>q \<cdot> \<gamma>\<rangle>" .
    from q have qg: "q \<cdot> \<gamma> \<notin> NF_terms Q" unfolding NF_def by force
    show False
    proof (cases C)
      case (More ff bef D aft)
      hence "Fun g ts \<cdot> ?\<sigma> \<rhd> q \<cdot> \<gamma>" using match l by auto
      with NF have "q \<cdot> \<gamma> \<in> NF_terms Q" by auto
      with qg show False by simp
    next
      case Hole
      with match have match: "?l\<cdot>?\<sigma> = q \<cdot> \<gamma>" by auto
      show ?thesis 
      proof (cases q)
        case (Fun ff qq)
        from match[unfolded l this] have ff: "ff = shp g" and qq: "length ts = length qq" using map_eq_imp_length_eq[of _ ts _ qq] by auto
        from ndef[OF qq[symmetric]] q[unfolded Fun ff]  
        show ?thesis by auto
      next
        case (Var x)
        from q[unfolded Var] wfQ show False by auto
      qed
    qed 
  qed
  hence NF: "?l \<cdot> ?\<sigma> \<in> NF_terms ?Q" by auto 
  have nfs: "NF_subst True (?l,?r) \<sigma> Q"
    unfolding NF_subst_def
  proof (intro impI subsetI)
    fix t
    assume t: "t \<in> \<sigma> ` vars_rule (?l,?r)"
    then obtain x where x: "x \<in> vars_rule (?l,?r)" and t: "t = \<sigma> x" by auto
    from x have "x \<in> vars_term ?l \<or> x \<in> vars_term ?r" unfolding vars_rule_def by auto
    hence "x \<in> vars_term ?l \<or> \<sigma> x \<in> NF_terms Q"
    proof 
      assume "x \<in> vars_term ?r"
      hence x: "x \<in> vars_term u" unfolding `u = Fun f us` by auto
      with supteq_imp_vars_subset[OF `r \<unrhd> u`] have x: "x \<in> vars_term r" by auto
      from wf show ?thesis
      proof
        assume "wwf_qtrs Q R"
        with lr x have "x \<in> vars_term l" unfolding wwf_qtrs_def applicable_rules_def by auto
        with `l = Fun g ts`
        show ?thesis by simp
      next
        assume "wwf_inn_qtrs nfs Q R"
        from this[unfolded wwf_inn_qtrs_def] have nfs by auto
        from nfs[unfolded NF_subst_def] vars_rule_def `(l,r) \<in> R` `nfs` x
        have "\<sigma> x \<in> NF_terms Q" by force
        thus ?thesis by simp
      qed
    qed simp
    hence "\<sigma> x \<in> NF_terms Q"
    proof
      assume "x \<in> vars_term ?l"
      hence "Var x \<unlhd> ?l" by auto
      hence "Var x \<cdot> \<sigma> \<unlhd> ?l \<cdot> \<sigma>" by auto
      from NF_subterm[OF NF this] show ?thesis by simp
    qed
    thus "t \<in> NF_terms Q" unfolding t .
  qed
  from NF nfs `(?l,?r) \<in> DP shp ?U` and `?r\<cdot>?\<sigma> = sharp_term shp ?u`
    and `?u \<in> Tinf ?qr` and `(sharp_term shp s, ?l\<cdot>?\<sigma>) \<in> ?Iqr^*`
  show ?thesis by blast
qed
  

lemma someI_ex4:
  "\<exists>w x y z. P (w, x, y, z) \<Longrightarrow> P (SOME (w, x, y, z). P (w, x, y, z))"
by (simp add: some_eq_ex)

fun
  next4 :: "('f \<Rightarrow> 'f) \<Rightarrow> bool \<Rightarrow> ('f,'v)term set \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)term \<Rightarrow>
    ('f,'v)term \<times> ('f,'v)term \<times> ('f,'v)term \<times> ('f,'v)subst"
where
  "next4 shp nfs Q R s = (SOME (l, r, u, \<sigma>).
    (l, r) \<in> DP shp (applicable_rules Q R) \<and> r \<cdot> \<sigma> = sharp_term shp u \<and> u \<in> Tinf (qrstep nfs Q R)
    \<and> (sharp_term shp s, l \<cdot> \<sigma>) \<in> (nrqrstep nfs Q R)^*
    \<and> l \<cdot> \<sigma> \<in> NF_terms Q \<and> NF_subst True (l,r) \<sigma> Q)"

fun
  concrete_seq :: "('f \<Rightarrow> 'f) \<Rightarrow> bool \<Rightarrow> ('f, 'v) term set \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) term \<Rightarrow> nat \<Rightarrow>
    ('f, 'v) term \<times> ('f, 'v) term \<times> ('f, 'v) term \<times> ('f, 'v) term \<times> ('f, 'v) subst"
where
  "concrete_seq shp nfs Q R s 0 = (s, next4 shp nfs Q R s)" |
  "concrete_seq shp nfs Q R r (Suc n) = (
    let ( _, _, _, s, _) = concrete_seq shp nfs Q R r n in
    (s, next4 shp nfs Q R s))"

theorem not_SN_imp_min_ichain:
  assumes "wf_qtrs nfs Q R" and "\<not> SN (qrstep nfs Q R)" 
  and ndef: "\<And> f n ss. defined (applicable_rules Q R) (f,n) \<Longrightarrow> length ss = n \<Longrightarrow> Fun (shp f) ss \<notin> Q"
  and m: "m = (\<forall> f n. defined (applicable_rules Q R) (f,n) \<longrightarrow> \<not> defined (applicable_rules Q R) (shp f,n))"
  shows "\<exists>s t \<sigma>. min_ichain (initial_dpp shp nfs m Q R) s t \<sigma>"
proof -
  let ?qr = "qrstep nfs Q R"
  let ?Q = "Q" and ?R = "R"
  let ?Iqr = "nrqrstep nfs ?Q ?R"
  from `\<not>(SN(qrstep nfs Q R))` obtain s' where "\<not> (SN_on ?qr {s'})"
    unfolding SN_on_def by auto
  then obtain s where "s' \<unrhd>  s" and "s \<in> Tinf ?qr" using not_SN_imp_subt_Tinf by auto
  from `wf_qtrs nfs Q R`
  have wwf: "wwf_qtrs Q R \<or> wwf_inn_qtrs nfs Q R" unfolding wf_qtrs_def by auto
  from wwf_inn_qtrs_imp_left_fun[OF wwf]
  have nvar: "\<forall> (l,r) \<in> R. is_Fun l" by auto    
  let ?I = "\<lambda>i. fst(concrete_seq shp nfs Q R s i)"
  let ?S = "\<lambda>i. fst(snd(concrete_seq shp nfs Q R s i))"
  let ?T = "\<lambda>i. fst(snd(snd(concrete_seq shp nfs Q R s i)))"
  let ?\<sigma> = "\<lambda>i. snd(snd(snd(snd(concrete_seq shp nfs Q R s i))))"
  let ?P = "\<lambda>s l r u \<sigma>. (l,r) \<in> DP shp (applicable_rules Q R) \<and> r\<cdot>\<sigma> = sharp_term shp u
    \<and> u \<in> Tinf ?qr \<and> (sharp_term shp s,l\<cdot>\<sigma>) \<in> ?Iqr^* \<and> l \<cdot> \<sigma> \<in> NF_terms ?Q \<and> NF_subst True (l,r) \<sigma> Q"
  let ?P' =  "\<lambda>(l,r,u,\<sigma>). ?P s l r u \<sigma>"
  have main: "\<forall>i. ?P (?I i) (?S i) (?T i) (?I(Suc i)) (?\<sigma> i)"
  proof
    fix i show "?P (?I i) (?S i) (?T i) (?I(Suc i)) (?\<sigma> i)"
    proof (induct i)
      case 0 
      have "\<exists>l r u \<sigma>. ?P'(l,r,u,\<sigma>)"
        using Tinf_lemma_DP[OF `wf_qtrs nfs Q R` `s \<in> Tinf ?qr` ndef] unfolding split by auto
      hence "?P'(next4 shp nfs Q R s)" unfolding next4.simps using someI_ex4[where P = "?P'"] by best
      have ini: "?I 0 = s" unfolding concrete_seq.simps by simp
      have s: "?S 0 = fst(next4 shp nfs Q R s)" unfolding concrete_seq.simps by simp
      have t: "?T 0 = fst(snd(next4 shp nfs Q R s))" unfolding concrete_seq.simps by simp
      have sub: "?\<sigma> 0 = snd(snd(snd(next4 shp nfs Q R s)))"
        unfolding concrete_seq.simps Let_def split_def by auto
      have nxt: "?I(Suc 0) = fst(snd(snd(next4 shp nfs Q R s)))"
        unfolding concrete_seq.simps Let_def split_def by auto
      show ?case unfolding ini s t sub nxt using `?P'(next4 shp nfs Q R s)` by auto
    next
      case (Suc i)
      let ?P'' = "\<lambda>(w,x,y,z). ?P (?I(Suc i)) w x y z"
      from Suc have "?I (Suc i) \<in> Tinf ?qr" by simp
      have "\<exists>l r u \<sigma>. ?P''(l,r,u,\<sigma>)"
        using Tinf_lemma_DP[OF `wf_qtrs nfs Q R` `?I (Suc i) \<in> Tinf ?qr` ndef] by auto
      hence IH: "?P''(next4 shp nfs Q R (?I(Suc i)))"
        unfolding next4.simps using someI_ex4[where P = "?P''"] by simp
      have s: "?S(Suc i) = fst(next4 shp nfs Q R (?I(Suc i)))"
        unfolding concrete_seq.simps Let_def split_def by simp
      have t: "?T(Suc i) = fst(snd(next4 shp nfs Q R (?I(Suc i))))"
        unfolding concrete_seq.simps Let_def split_def by simp
      have sub: "?\<sigma>(Suc i) = snd(snd(snd(next4 shp nfs Q R (?I(Suc i)))))"
        unfolding concrete_seq.simps Let_def split_def by simp
      have nxt: "?I(Suc(Suc i)) = fst(snd(snd(next4 shp nfs Q R (?I(Suc i)))))"
        unfolding concrete_seq.simps Let_def split_def by simp
      show ?case unfolding s t sub nxt using IH by auto
    qed
  qed
  from main have P: "\<forall>i. (?S i,?T i) \<in> DP shp (applicable_rules Q R)" by simp
  from main have NF: "\<And>i. (?S i \<cdot> ?\<sigma> i) \<in> NF_terms ?Q" by auto
  from main have nfs: "\<And>i. NF_subst nfs (?S i, ?T i) (?\<sigma> i) Q" unfolding NF_subst_def by blast
  from main have eq: "\<And>i. sharp_term shp (?I(Suc i)) = ?T i \<cdot> ?\<sigma> i" by auto
  have R: "\<forall>i. (?T i \<cdot> ?\<sigma> i,?S(Suc i) \<cdot> ?\<sigma>(Suc i)) \<in> ?Iqr^*"
  proof
    fix i show "(?T i \<cdot> ?\<sigma> i,?S(Suc i)\<cdot>?\<sigma>(Suc i)) \<in> ?Iqr^*"
    proof (cases i)
      case 0
      from main have first: "?T 0 \<cdot> ?\<sigma> 0 = sharp_term shp (?I(Suc 0))" by best
      from main have "(sharp_term shp (?I(Suc 0)),?S(Suc 0)\<cdot>?\<sigma>(Suc 0)) \<in> ?Iqr^*" by best
      thus ?thesis unfolding 0 unfolding first by simp
    next
      case (Suc i) show ?thesis unfolding eq[symmetric] using main by best
    qed
  qed
  have Tinf: "\<forall>i. ?I i \<in> Tinf ?qr"
  proof
    fix i show "?I i \<in> Tinf ?qr" using main by (induct i) (auto simp: `s \<in> Tinf ?qr`)
  qed
  {
    fix i 
    from Tinf have Tinf: "?I (Suc i) \<in> Tinf ?qr" by best
    from main have "(?S i, ?T i) \<in> DP shp (applicable_rules Q R)" by auto
    then obtain f ts where ti: "?T i = Fun (shp f) ts" and "defined (applicable_rules Q R) (f,length ts)" unfolding DP_def by auto
    then obtain ts where id: "?T i \<cdot> ?\<sigma> i = Fun (shp f) ts" and d: "defined (applicable_rules Q R) (f,length ts)"  by auto
    from eq id have eq: "sharp_term shp (?I(Suc i)) = Fun (shp f) ts" by simp
    from eq obtain g where "?I (Suc i) = Fun g ts" by (cases "?I (Suc i)", auto)
    from Tinf[unfolded this] have "Fun g ts \<in> Tinf ?qr" .
    from this[unfolded Tinf_def] have SN: "\<And> t. t \<in> set ts \<Longrightarrow> SN_on ?qr {t}" by auto
    hence "\<And> t. t \<in> set (args (?T i \<cdot> ?\<sigma> i)) \<Longrightarrow> SN_on ?qr {t}" unfolding id by simp
  } note SN_args = this
  have RR: "?Iqr^* \<subseteq> ?qr^*" by (rule rtrancl_mono, unfold qrstep_iff_rqrstep_or_nrqrstep, auto)
  from R have R: "\<forall> i. (?T i \<cdot> ?\<sigma> i,?S(Suc i) \<cdot> ?\<sigma>(Suc i)) \<in> ?qr^*" using RR by blast
  note idd = initial_dpp.simps
  from P and R and NF nfs have ichain: "ichain (initial_dpp shp nfs m Q R) ?S ?T ?\<sigma>" 
    by (auto simp: ichain.simps)
  show ?thesis
  proof (intro exI, unfold idd min_ichain.simps, unfold idd[symmetric], rule conjI[OF ichain], rule impI)
    assume m
    hence ndef2: "\<forall> f n. defined (applicable_rules Q R) (f,n) \<longrightarrow> \<not> defined (applicable_rules Q R) (shp f, n)"
      unfolding m .
    have SN_t: "\<forall>i. SN_on ?qr {?T i \<cdot> ?\<sigma> i}"
    proof
      fix i
      from Tinf have Tinf: "?I (Suc i) \<in> Tinf ?qr" by best
      from main have "(?S i, ?T i) \<in> DP shp (applicable_rules Q R)" by auto
      then obtain f ts where "?T i = Fun (shp f) ts" and "defined (applicable_rules Q R) (f,length ts)" unfolding DP_def by auto
      then obtain ts where id: "?T i \<cdot> ?\<sigma> i = Fun (shp f) ts" and d: "defined (applicable_rules Q R) (f,length ts)"  by auto
      show "SN_on ?qr {?T i \<cdot> ?\<sigma> i}" unfolding id 
      proof (rule SN_args_imp_SN[OF _ nvar ndef2[THEN spec, THEN spec, THEN mp[OF _ d]]])
        fix t
        assume t: "t \<in> set ts"
        with SN_args[of t i, unfolded id]
        show "SN_on ?qr {t}" by auto
      qed
    qed
    thus "minimal_cond nfs Q ({} \<union> R) ?S ?T ?\<sigma>"
      unfolding minimal_cond_def by simp
  qed
qed

hide_const next4 concrete_seq

corollary not_SN_imp_ichain:
  assumes prem: "wf_qtrs nfs Q R" "\<not> SN (qrstep nfs Q R)" 
  "\<And> f n ss. defined (applicable_rules Q R) (f,n) \<Longrightarrow> length ss = n \<Longrightarrow> Fun (shp f) ss \<notin> Q"
  shows "\<exists>s t \<sigma>. ichain (initial_dpp shp nfs m Q R) s t \<sigma>"
proof -
  obtain mm where mm: "mm = (\<forall>f n. defined (applicable_rules Q R) (f, n) \<longrightarrow>
                    \<not> defined (applicable_rules Q R) (shp (f), n))" by auto
  from not_SN_imp_min_ichain[OF prem mm] 
  have "\<exists>s t \<sigma>. min_ichain (initial_dpp shp nfs mm Q R) s t \<sigma>" .
  thus ?thesis unfolding initial_dpp.simps min_ichain.simps ichain.simps by blast
qed


corollary not_SN_imp_ichain_rstep:
  assumes wf: "wf_trs R" and nSN: "\<not> SN (rstep R)" 
  shows "\<exists> s t \<sigma>. ichain (nfs,m,DP shp R, {}, {}, {}, R) s t \<sigma>"
proof -
  let ?DP = "DP shp (applicable_rules {} R)"
  have U: "applicable_rules {} R = R" unfolding applicable_rules_def applicable_rule_def by auto
  from wf U have wf: "wf_qtrs nfs {} R" unfolding wf_qtrs_def wwf_qtrs_wf_trs by auto
  from nSN have nSN: "\<not> SN (qrstep nfs {} R)" by auto
  have id: "initial_dpp shp nfs m {} R = (nfs,m,DP shp R, {}, {}, {}, R)" by (simp add: applicable_rules_def applicable_rule_def)
  from not_SN_imp_ichain[OF wf nSN, of shp m, unfolded id U] show ?thesis by auto
qed

lemma sharp_term_id: "sharp_term id t = t"
  by (induct t, auto)

(* completeness lemma *)
lemma SN_imp_finite_dpp_simple: fixes R :: "('f,'v)trs"
  assumes SN: "SN (qrstep nfs Q R)"  
  and shp: "\<And> f n. unshp \<noteq> id \<Longrightarrow> (f,n) \<in> D \<Longrightarrow> \<not> defined R (shp f,n)" 
  and unshp: "\<And> f n. (f,n) \<in> D \<Longrightarrow> unshp (shp f) = f"
  and D: "\<And> fn. defined R fn \<Longrightarrow> fn \<in> D"
  and nfs: "nfs \<Longrightarrow> \<forall> l r. (l,r) \<in> R \<longrightarrow> is_Fun l"
  and nnfs: "\<not> nfs" (* TODO: delete nnfs *)
  shows "finite_dpp (nfs,m,DP_simple shp D R,{},Q,{},R)"
proof (rule ccontr)
  let ?DP = "DP_simple shp D R"
  let ?s = "sharp_term shp :: ('f,'v)term \<Rightarrow> ('f,'v)term"
  let ?u = "sharp_term unshp :: ('f,'v)term \<Rightarrow> ('f,'v)term"
  let ?r = "qrstep nfs Q R \<union> {\<rhd>}"
  assume "\<not> ?thesis"
  then obtain s t \<sigma> where "ichain (nfs,m,?DP,{},Q,{},R) s t \<sigma>" 
    unfolding finite_dpp_def by auto
  note ichain = this[unfolded ichain.simps]
  {
    fix l r x
    assume lr: "(l,r) \<in> R" and l: "l = Var x" 
    have False
    proof (cases nfs)
      case False
      from left_Var_imp_not_SN_qrstep[OF lr[unfolded l] False, of Q]
        SN show False unfolding SN_def by auto
    next
      case True
      with nfs[OF this] lr l show False by auto
    qed
  } note nvar = this
  hence varcond: "\<forall> (l,r) \<in> R. is_Fun l" by force
  {
    fix i
    from ichain have "(s i, t i) \<in> ?DP" by auto
    from this[unfolded DP_simple_def] obtain l r h us where s: "s i = ?s l"
      and t: "t i = ?s (Fun h us)" and lr: "(l,r) \<in> R" and ru: "r \<unrhd> Fun h us"
      and h: "(h,length us) \<in> D" by auto
    from nvar[OF lr]
    obtain f ls where l: "l = Fun f ls" by (cases l, auto)
    from lr[unfolded l] have f: "defined R (f, length ls)" unfolding defined_def by auto
    from ru obtain C where r: "r = C\<langle>Fun h us\<rangle>" by auto
    have si: "?u (s i \<cdot> \<sigma> i) =  l \<cdot> \<sigma> i" unfolding s l using unshp[OF D[OF f]]
      by simp
    from ichain have "s i \<cdot> \<sigma> i \<in> NF_terms Q" and nfs: "NF_subst nfs (s i, t i) (\<sigma> i) Q"
      by auto
    from NF_imp_subt_NF[OF this(1)]
    have "\<forall> t \<in> set (args (s i \<cdot> \<sigma> i)). t \<in> NF_terms Q"
      unfolding NF_terms_args_conv .
    hence "\<forall> t \<in> set (args (l \<cdot> \<sigma> i)). t \<in> NF_terms Q" unfolding s l by simp
    hence NF: "\<forall> t \<lhd> l \<cdot> \<sigma> i. t \<in> NF_terms Q" unfolding NF_terms_args_conv .
    from nfs nnfs have nfs: "NF_subst nfs (l,r) (\<sigma> i) Q" by simp
    have step: "(l \<cdot> \<sigma> i, r \<cdot> \<sigma> i) \<in> qrstep nfs Q R"
      by (rule qrstepI[OF _ lr, of "\<sigma> i" _ _ \<box>, OF NF], auto simp: nfs)
    have supt: "(r \<cdot> \<sigma> i, Fun h us \<cdot> \<sigma> i) \<in> {\<rhd>}\<^sup>=" using supteq_stable[OF ru, of "\<sigma> i"] unfolding supteq_supt_set_conv .
    from ichain have steps: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (qrstep nfs Q R)^*" by auto    
    have "(?u (t i \<cdot> \<sigma> i), ?u (s (Suc i) \<cdot> \<sigma> (Suc i))) \<in> (qrstep nfs Q R)^*" 
    proof (cases "unshp = id")
      case False
      have steps: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (nrqrstep nfs Q R)^*"
      proof (rule qrsteps_imp_nrqrsteps[OF varcond ndef_applicable_rules steps])
        show "\<not> defined R (the (root (t i \<cdot> \<sigma> i)))"
          unfolding t using shp False h by auto
      qed
      from nrqrsteps_imp_sharp_qrsteps[OF steps, of unshp]
      show ?thesis
        unfolding qrstep_iff_rqrstep_or_nrqrstep by regexp
    next
      case True
      from steps show ?thesis unfolding True sharp_term_id .
    qed
    hence steps: "(Fun h us \<cdot> \<sigma> i, ?u (s (Suc i) \<cdot> \<sigma> (Suc i))) \<in> (qrstep nfs Q R)^*"
      unfolding t using unshp[OF h] by simp
    from si step supt steps have steps: "(?u (s i \<cdot> \<sigma> i), ?u (s (Suc i) \<cdot> \<sigma> (Suc i))) \<in>
      qrstep nfs Q R O {\<rhd>}^= O (qrstep nfs Q R)^*" by auto
    have "(?u (s i \<cdot> \<sigma> i), ?u (s (Suc i) \<cdot> \<sigma> (Suc i))) \<in> ?r^+"
      by (rule set_mp[OF _ steps], regexp)
  } note main = this
  obtain u where u: "u = (\<lambda> i. ?u (s i \<cdot> \<sigma> i))" by auto
  from main have "\<And> i. (u i, u (Suc i)) \<in> ?r^+" unfolding u .
  hence "\<not> SN (?r^+)" by auto
  hence "\<not> SN ?r" using SN_imp_SN_trancl[of ?r] by auto
  thus False using SN_imp_SN_union_supt[OF SN ctxt_closed_qrstep] ..
qed

lemma SN_imp_finite_dpp: fixes R :: "('f,'v)trs"
  assumes SN: "SN (qrstep nfs Q R)"  
  and shp: "\<And> f n. unshp \<noteq> id \<Longrightarrow> defined R (f,n) \<Longrightarrow> \<not> defined R (shp f,n)" 
  and unshp: "\<And> f n. defined R (f,n) \<Longrightarrow> unshp (shp f) = f"
  and nfs: "nfs \<Longrightarrow> \<forall> l r. (l,r) \<in> R \<longrightarrow> is_Fun l"
  and nnfs: "\<not> nfs" (* TODO: delete nnfs *)
  shows "finite_dpp (nfs,m,DP shp R,{},Q,{},R)"
  by (rule finite_dpp_mono[OF SN_imp_finite_dpp_simple[OF SN shp unshp _ nfs nnfs, of "{fn . defined R fn}"]],
    auto simp: DP_subset_DP_simple)

end

